import ast
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.fernet import Fernet, InvalidToken
from base64 import urlsafe_b64encode
from .exceptions import PasswordError, NoTargetFoundError
from .etypes import DecryptedSecret, EncryptedSecret
import logging

logger = logging.getLogger(__name__)
logger.level = logging.ERROR

DEFAULT_SALT = (
    b"\x89\x7f\xc8\x93\x1d\xb8\xd3\xea\x16\x91\x08\xad\x15\x96\x1b\xef"  # A fixed salt
)


def derive_key(password: str, salt: str = None) -> bytes:
    if not salt:
        salt = DEFAULT_SALT

    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend(),
    )
    return urlsafe_b64encode(kdf.derive(password.encode()))


def encrypt_value(key: bytes, secure_value: str) -> str:
    cipher_suite = Fernet(key)
    return cipher_suite.encrypt(secure_value.value.encode()).decode()


def decrypt_value(key: bytes, encrypted_value: str) -> str:
    cipher_suite = Fernet(key)
    try:
        return cipher_suite.decrypt(encrypted_value.value.encode()).decode()
    except InvalidToken:
        raise PasswordError("The password is incorrect or the salt does not match.")


def process_file(
    file_path: str,
    key: bytes,
    encrypt: bool = True,
    dry_run: bool = False,
    verbose: bool = False,
):
    if verbose:
        logger.level = logging.INFO

    with open(file_path, "r+") as file:
        source = file.read()
        logger.info(f"Processing: {file_path}")
        tree = ast.parse(source, filename=file_path)
        modified_source = source
        found = False

    for node in ast.walk(tree):
        if isinstance(node, ast.AnnAssign) and isinstance(node.value, ast.Constant):
            if isinstance(node.annotation, ast.Name):
                if encrypt and node.annotation.id == str(DecryptedSecret.__name__):
                    encrypted_value = encrypt_value(key, node.value)
                    modified_source = modified_source.replace(
                        node.value.value, encrypted_value
                    )
                    modified_source = modified_source.replace(
                        str(DecryptedSecret.__name__), str(EncryptedSecret.__name__)
                    )
                    found = True
                elif not encrypt and node.annotation.id == str(EncryptedSecret.__name__):
                    decrypted_value = decrypt_value(key, node.value)
                    modified_source = modified_source.replace(
                        node.value.value, decrypted_value
                    )
                    modified_source = modified_source.replace(
                        str(EncryptedSecret.__name__), str(DecryptedSecret.__name__)
                    )
                    found = True

    if not found:
        if encrypt:
            raise NoTargetFoundError(f"No {str(DecryptedSecret.__name__)} hiints found for encryption.")
        else:
            raise NoTargetFoundError(f"No {str(EncryptedSecret.__name__)} hints found for decryption.")

    if modified_source != source:
        if not dry_run:
            with open(file_path, "w") as file:
                file.write(modified_source)
                logger.info(f"Processed: {file_path}")
        else:
            logger.info(f"[DRY RUN] Processed file: {file_path}, did not save")
            return modified_source


def encrypt_file(file_path, password, salt=None, dryrun=False, verbose=False):
    if verbose:
        logger.level = logging.INFO

    key = derive_key(password, salt=salt)
    return process_file(file_path, key, encrypt=True, dry_run=dryrun, verbose=verbose)


def decrypt_file(file_path, password, salt=None, dryrun=False, verbose=False):
    if verbose:
        logger.level = logging.INFO

    key = derive_key(password, salt=salt)
    return process_file(file_path, key, encrypt=False, dry_run=dryrun, verbose=verbose)



import os
import logging
import click

from etypes.processor import derive_key, encrypt_file, decrypt_file

logger = logging.getLogger(__name__)
logger.level = logging.ERROR


@click.group()
def main():
    pass


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("file_path")
@click.option("-p", "--password", help="password for specified file_path")
@click.option("-pf", "--password-file", help="password file for specified file_path")
@click.option("-s", "--salt", help="optional salt for specified file_path")
@click.option("--dryrun", is_flag=True, default=False, help="Dry run verbose mode")
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Enables verbose mode"
)
def encrypt(file_path, password, password_file, salt, dryrun, verbose):

    if verbose:
        logger.level = logging.INFO

    if password_file:
        if not os.path.isfile(password_file):
            raise FileNotFoundError(password_file)
        with open(password_file, "r") as file:
            password = file.read().strip()
        logger.info(f"Loaded password file {password_file}")

    if not os.path.isfile(file_path):
        raise FileNotFoundError(file_path)

    encrypt_file(file_path, password, salt=salt, dryrun=dryrun, verbose=verbose)
    logger.info(f"Encrypted file {password_file}")


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("file_path")
@click.option("-p", "--password", help="password for specified file_path")
@click.option("-pf", "--password-file", help="password file for specified file_path")
@click.option("-s", "--salt", help="optional salt for specified file_path")
@click.option("--dryrun", is_flag=True, default=False, help="Dry run verbose mode")
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Enables verbose mode"
)
def decrypt(file_path, password, password_file, salt, dryrun, verbose):

    if verbose:
        logger.level = logging.INFO

    if password_file:
        if not os.path.isfile(password_file):
            raise FileNotFoundError(password_file)
        with open(password_file, "r") as file:
            password = file.read().strip()
        logger.info(f"Loaded password file {password_file}")

    if not os.path.isfile(file_path):
        raise FileNotFoundError(file_path)

    decrypt_file(file_path, password, salt=salt, dryrun=dryrun, verbose=verbose)
    logger.info(f"Decrypted file {password_file}")


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("file")
def dump():
    pass


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("config")
def verify():
    pass


main.add_command(encrypt)
main.add_command(decrypt)
main.add_command(verify)
main.add_command(dump)



import os
import tempfile
import shutil


def create_temporary_copy(src_file_name, preserve_extension=False):
    """
    Copies the source file into a temporary file.
    Returns a _TemporaryFileWrapper, whose destructor deletes the temp file
    (i.e. the temp file is deleted when the object goes out of scope).
    """
    tf_suffix = ""
    if preserve_extension:
        _, tf_suffix = os.path.splitext(src_file_name)
    tf = tempfile.NamedTemporaryFile(suffix=tf_suffix)
    shutil.copy2(src_file_name, tf.name)
    return tf



import os
import ast

from typing import get_type_hints

from etypes import EncryptedSecret
from etypes.processor import decrypt_value, derive_key
from etypes.exceptions import NoPasswordProvidedError, NoPasswordFileProvidedError

class AutoLoader:
    password = None
    password_file = None

    def process(self, value):
        key = derive_key(self.password)
        return decrypt_value(key, ast.Constant(value))

    def walk(self, node, annotations):
        if isinstance(node, dict):
            for key, value in node.items():
                if key[0] == "_":
                    # ignore private and builtins
                    continue

                if isinstance(value, dict):
                    node[key] = self.walk(value, annotations)
                if isinstance(value, list):
                    for v in value:
                        if isinstance(value, EncryptedSecret):
                            node[key] = self.process(v)

                if key in annotations:
                    if annotations[key] == EncryptedSecret:
                        node[key] = self.process(value)

        return node

    
    def __init__(self, locals, password_var_name=None, password_file=None):

        if not password_file and not password_var_name:
            raise Exception("You must specify a password_var_name or password_file")
        
        if password_var_name and not password_file:
            self.password = os.environ.get(password_var_name, None)
            if not self.password:
                raise NoPasswordProvidedError(
                    f"No password found in environment variable: {password_var_name}"
                )

        if not password_var_name and password_file:
            if not os.path.isfile(password_file):
                raise NoPasswordFileProvidedError(f"Could not find {password_file}")
            password = open(password_file,'r').read()
            if not password:
                raise NoPasswordProvidedError(f'No password found in {password_file}')
            self.password = password

        annotations = locals["__annotations__"]
        self.walk(locals, annotations)



class EncryptedSecret(str):
    """Encrypted version of a DecryptedSecret"""
    pass

class DecryptedSecret(str):
    """Decrypted version of a EncryptedSecret"""
    pass

class EnvString(str):
    pass



#################################################
#   [2024] Dan Brosnan dpjbrosnan@gmail.com     #
#################################################

# from __future__ import absolute_import, division, print_function
import os

__metaclass__ = type

DOCUMENTATION = """
    vars: etypes
    version_added: "2.10"
    short_description: Load custom host vars from an encrypted python file
    description: Load custom host vars from an encrypted python file
    extends_documentation_fragment:
      - vars_plugin_staging
"""


try:
    import ansible
except ImportError:
    raise Exception("Requires ansible to be installed")

from ansible.plugins.vars import BaseVarsPlugin
from ansible.utils.vars import combine_vars

class VarsModule(BaseVarsPlugin):
    """Allows decryption of a etypes encrypted python file for use as ansible vars"""
    allow_extras = True
    REQUIRES_ENABLED = False
    is_stateless = True

    def load_found_files(self, loader, data, found_files):
        for found in found_files:
            new_data = loader.load_from_file(found, cache="all", unsafe=True)
            if new_data:  # ignore empty files
                data = combine_vars(data, new_data)
        return data

    def get_vars(self, loader, path, entities, cache=True):
        """parses the inventory file"""

        etypes_file = os.environ.get('ETYPES_FILE', None)
        if not etypes_file:
            raise Exception("ETYPES_FILE not specified")
        
        etypes_password = os.environ.get('ETYPES_PASSWORD', None)
        if not etypes_password:
            raise Exception("ETYPES_PASSWORD not specified")
        data = {}
        with open(etypes_file, 'r') as fp:
            source = fp.read()
            _locals = locals()
            exec(source, globals(), _locals)        
            data = dict(locals())

        remove = ['AutoLoader', 'EncryptedSecret', 'DecryptedSecret', '__annotations__', '_locals', 'cache', 'data', 'entities', 'etypes_file', 'etypes_password', 'fp', 'loader', 'path', 'self', 'source']
        for r in remove:
            try:
                del data[r]
            except KeyError:
                pass
        return data


class EncryptionError(Exception):
    """Base class for encryption-related errors."""

    pass


class DecryptionError(Exception):
    """Base class for decryption-related errors."""

    pass


class PasswordError(EncryptionError, DecryptionError):
    """Exception raised for errors involving the encryption or decryption key."""

    pass


class NoTargetFoundError(Exception):
    """Exception raised when no target strings are found for processing."""
    pass


class NoPasswordProvidedError(PasswordError):
    """Exception raised when no password is provided."""
    pass

class NoPasswordFileProvidedError(FileNotFoundError):
    """Exception raised when no password file is found."""
    pass



import os
from .exceptions import NoPasswordProvidedError
from .processor import decrypt_file
from .utils import create_temporary_copy


def dynamic_import(globals, file_path, password=None, password_file=None, salt=None):

    if not password and not password_file:
        raise NoPasswordProvidedError(f"No password or password_file was provided")

    if password_file:
        if not os.path.isfile(password_file):
            raise FileNotFoundError(f"Could not find password_file {password_file}")
        with open(password_file, "r") as file:
            password = file.read()

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Could not find {file_path}")

    modified_source = None
    with create_temporary_copy(file_path) as fp:
        temp_file_path = fp.name
        decrypt_file(temp_file_path, password=password)
        modified_source = open(temp_file_path, "r").read()

    exec(modified_source, globals)
    return globals
